/*--------------------------- MegaWave2 Module -----------------------------*/
/* mwcommand 
  name = {fgrain}; 
  version = {"1.2"}; 
  author = {"Pascal Monasse, Frederic Guichard, G. Facciolo"}; 
  function = {"Grain filter of an image"}; 
  usage = { 
    'a': [min_area=20]-> pMinArea   "Min area of grains we keep", 
    image_in -> pFloatImageInput    "Input fimage", 
    image_out <- pFloatImageOutput  "Output fimage" 
    }; 
*/ 
/*----------------------------------------------------------------------
 v1.2 (04/2007): simplified header (LM)
 v1.3 (2013): portable version (GF)
----------------------------------------------------------------------*/

#include "mw3.h" 
#include "iio.h"
#include "stdio.h"
#include "stdlib.h"

extern void flst();
extern void flst_reconstruct();




// ===================================================================
// WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- START HERE
//                                   (20.08.2015)

/* This removes the shapes from the tree associated to pFloatImageInput
that are too small (threshold *pMinArea). As a consequence all the remaining 
shapes of pFloatImageOutput are of area larger or equal than *pMinArea */ 

void mw_fgrain(Fimage pFloatImageInput,
               int *pMinArea1, Fimage pFloatImageOutput1,
               int *pMinArea2, Fimage pFloatImageOutput2,
               int *pMinArea3, Fimage pFloatImageOutput3,
               int *pMinArea4, Fimage pFloatImageOutput4,
               int *pMinArea5, Fimage pFloatImageOutput5,
               int *pMinArea6, Fimage pFloatImageOutput6,
               int *pMinArea7, Fimage pFloatImageOutput7,
               int *pMinArea8, Fimage pFloatImageOutput8,
               int *pMinArea9, Fimage pFloatImageOutput9 )
{
    int i,lev; // WARNING: MODIFICATION ON THIS LINE -- Charles (19.08.2015)
    Shapes pTree;
    int TabAreas[9] = {
        *pMinArea1,*pMinArea2,*pMinArea3,
        *pMinArea4,*pMinArea5,*pMinArea6,
        *pMinArea7,*pMinArea8,*pMinArea9};
    
    if(mw_change_fimage(pFloatImageOutput1, pFloatImageInput->nrow,
                        pFloatImageInput->ncol) == NULL)
        mwerror(FATAL, 1,
                "fgrain --> Not enough memory to allocate the output image");
    if((pTree = mw_new_shapes()) == NULL)
        mwerror(FATAL, 1,
                "fgrain --> Not enough memory to allocate the tree of shapes");
    
    
    printf("Computation of the flst of the image... ");
    
    /* Compute the Level Sets Transform of the input image */
    flst(NULL, pFloatImageInput, pTree, NULL, NULL);
    
    printf("ok\n");
    
    // ===================================================================
    // WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- START HERE
    //                                   (20.08.2015)
    
    printf("Nine applications of the Grain filter... ");
    
    // Loop over the 9 grain sizes
    for(lev=0;lev<9;lev++){
        
        // WARNING: MODIFICATIONS. Keep in mind that the areas have to be sorted
        // !!!                     from the smallest size to the biggest.
        
        /* Kill too small grains.
         Bound i>0 because it is forbidden to delete the root, at index 0 */
        for(i = pTree->nb_shapes-1; i > 0; i--)
            if(pTree->the_shapes[i].area < TabAreas[lev])
                pTree->the_shapes[i].removed = (char)1;
        
        /* Reconstruct in pFloatImageOutput the modified tree of shapes */
        switch (lev+1) {
            case 1:
                flst_reconstruct(pTree, pFloatImageOutput1);
                break;
            case 2:
                flst_reconstruct(pTree, pFloatImageOutput2);
                break;
            case 3:
                flst_reconstruct(pTree, pFloatImageOutput3);
                break;
            case 4:
                flst_reconstruct(pTree, pFloatImageOutput4);
                break;
            case 5:
                flst_reconstruct(pTree, pFloatImageOutput5);
                break;
            case 6:
                flst_reconstruct(pTree, pFloatImageOutput6);
                break;
            case 7:
                flst_reconstruct(pTree, pFloatImageOutput7);
                break;
            case 8:
                flst_reconstruct(pTree, pFloatImageOutput8);
                break;
            case 9:
                flst_reconstruct(pTree, pFloatImageOutput9);
                break;
                
            default:
                mwerror(FATAL, 1,
                        "fgrain --> 'lev' variable is not an integer between 0 and 8.");
                break;
        }
        
    }
    
    printf("ok\n");
    
  // WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- END HERE
  // ===================================================================
  
  
  mw_delete_shapes(pTree); 
} 

// WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- END HERE
// ===================================================================


//  Main function and input/output.

Fimage fimageread(char* name)
{
      int nx,ny,nc;
         float *fdisp = iio_read_image_float_split(name, &nx, &ny, &nc);
         Fimage out = mw_change_fimage(NULL,ny,nx);
         free(out->gray);
         out->gray=fdisp;
         return out;
}

void fimagewrite(char* name, Fimage i)
{
      iio_save_image_float_split(name, i->gray, i->ncol, i->nrow, 1);
}


// ===================================================================
// WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- START HERE

/*in and out must be allocated*/
void fgrain(float *in, int nx, int ny,
            int MinArea1, float *out1,
            int MinArea2, float *out2,
            int MinArea3, float *out3,
            int MinArea4, float *out4,
            int MinArea5, float *out5,
            int MinArea6, float *out6,
            int MinArea7, float *out7,
            int MinArea8, float *out8,
            int MinArea9, float *out9 ) {
    int i;
    Fimage mwin = mw_change_fimage(NULL,ny,nx);
    Fimage mwout1 = mw_new_fimage();
    Fimage mwout2 = mw_new_fimage();
    Fimage mwout3 = mw_new_fimage();
    Fimage mwout4 = mw_new_fimage();
    Fimage mwout5 = mw_new_fimage();
    Fimage mwout6 = mw_new_fimage();
    Fimage mwout7 = mw_new_fimage();
    Fimage mwout8 = mw_new_fimage();
    Fimage mwout9 = mw_new_fimage();
    
    for(i=0;i<nx*ny;i++) mwin->gray[i] = in[i];
    
    mw_fgrain(mwin,
              &MinArea1, mwout1,
              &MinArea2, mwout2,
              &MinArea3, mwout3,
              &MinArea4, mwout4,
              &MinArea5, mwout5,
              &MinArea6, mwout6,
              &MinArea7, mwout7,
              &MinArea8, mwout8,
              &MinArea9, mwout9 );
    
    
    for(i=0;i<nx*ny;i++){
        out1[i] = mwout1->gray[i];
        out2[i] = mwout2->gray[i];
        out3[i] = mwout3->gray[i];
        out4[i] = mwout4->gray[i];
        out5[i] = mwout5->gray[i];
        out6[i] = mwout6->gray[i];
        out7[i] = mwout7->gray[i];
        out8[i] = mwout8->gray[i];
        out9[i] = mwout9->gray[i];
    }
    
    mw_delete_fimage(mwin);
    mw_delete_fimage(mwout1);
    mw_delete_fimage(mwout2);
    mw_delete_fimage(mwout3);
    mw_delete_fimage(mwout4);
    mw_delete_fimage(mwout5);
    mw_delete_fimage(mwout6);
    mw_delete_fimage(mwout7);
    mw_delete_fimage(mwout8);
    mw_delete_fimage(mwout9);
}

// WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- END HERE
// ===================================================================




int main(int argc, char* argv[])
{
    // default parameters
    
    // ===================================================================
    // WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- START HERE
    
    int gsize1 =   9;
    int gsize2 =  36;
    int gsize3 =  81;
    int gsize4 = 144;
    int gsize5 = 225;
    int gsize6 = 324;
    int gsize7 = 441;
    int gsize8 = 576;
    int gsize9 = 729;
    
    // char name_BW_img[] = "input_gray.png";
    
    char name1[] = "grain1.png";
    char name2[] = "grain2.png";
    char name3[] = "grain3.png";
    char name4[] = "grain4.png";
    char name5[] = "grain5.png";
    char name6[] = "grain6.png";
    char name7[] = "grain7.png";
    char name8[] = "grain8.png";
    char name9[] = "grain9.png";
    
    
    if(argc<2) {
        fprintf(stderr, "Grain filter of an image\n");
        fprintf(stderr, "syntax: %s image_in [gsize1=9 gsize2=36 ... gsize9=729]\n", argv[0]);
        return 1;
    }
    if(argc<11 & argc != 2) {
        fprintf(stderr, "9 grain sizes are required.\n");
        fprintf(stderr, "syntax: %s image_in [gsize1=9 gsize2=36 ... gsize9=729]\n", argv[0]);
        return 1;
    }
    else {
        if(argc != 2) {
            gsize1 = atoi(argv[ 2]);
            gsize2 = atoi(argv[ 3]);
            gsize3 = atoi(argv[ 4]);
            gsize4 = atoi(argv[ 5]);
            gsize5 = atoi(argv[ 6]);
            gsize6 = atoi(argv[ 7]);
            gsize7 = atoi(argv[ 8]);
            gsize8 = atoi(argv[ 9]);
            gsize9 = atoi(argv[10]);
        }
        else {
            printf("Grain Filter with sizes 9, 36, 81, 144, 225, 324, 441, 576 and 729.\n");
        }
    }
    
    Fimage in = fimageread(argv[1]);
    
    Fimage out_niv1 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv2 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv3 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv4 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv5 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv6 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv7 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv8 = mw_change_fimage(NULL,in->nrow,in->ncol);
    Fimage out_niv9 = mw_change_fimage(NULL,in->nrow,in->ncol);
    
    fgrain(in->gray, in->ncol, in->nrow,
           gsize1, out_niv1->gray,
           gsize2, out_niv2->gray,
           gsize3, out_niv3->gray,
           gsize4, out_niv4->gray,
           gsize5, out_niv5->gray,
           gsize6, out_niv6->gray,
           gsize7, out_niv7->gray,
           gsize8, out_niv8->gray,
           gsize9, out_niv9->gray);
    
    printf("Multigrain: Writing the 9 filtered images... ");
    
    // fimagewrite(name_BW_img, in);
    
    fimagewrite(name1, out_niv1);
    fimagewrite(name2, out_niv2);
    fimagewrite(name3, out_niv3);
    fimagewrite(name4, out_niv4);
    fimagewrite(name5, out_niv5);
    fimagewrite(name6, out_niv6);
    fimagewrite(name7, out_niv7);
    fimagewrite(name8, out_niv8);
    fimagewrite(name9, out_niv9);
    
    printf("ok\n");
    
    // WARNING: MODIFICATIONS -- CHARLES (19.08.2015) -- END HERE
    // ===================================================================
    
  return 0;
}
